from ._base_brute_force._base_brute_force import BruteForce

#NOTE: C Version looks like Python version wich is normal, because both are pretty much the same.
#NOTE: We can make a brute force with clusters maybe works

__all__ = ["BruteForce"]
